# Wireless configuration script
# Version 0.4.7
# Copyright (c) 2004 Roy "UberLord" Marples (uberlord@rsm.demon.co.uk)
# Released under the GPL version 2.0 or later
# http://www.gnu.org/licenses/gpl.html
#
# This is primarily tailored for use with Gentoo Linux >=sys-apps/baselayout-1.8.11
#
# ${devnull} is set by the calling script

source /etc/conf.d/wireless

wireless_checkconfig() {
	if [ ! -x "${IWCONFIG}" ]; then
		eerror "For Wireless (802.11) support, emerge net-wireless/wireless-tools"
		return 1
	fi

	# The Gentoo net.eth0 script should really check this
	${IFCONFIG} ${IFACE} &>${devnull} || {
		eerror "${IFACE} does not exist"
		return 1
	}
}

wireless_setup_programs() {
	IFCONFIG=`which ifconfig 2>${devnull}`
	IWCONFIG=`which iwconfig 2>${devnull}`
	IWLIST=`which iwlist 2>${devnull}`
	IWPRIV=`which iwpriv 2>${devnull}`
	IWGETID=`which iwgetid 2>${devnull}`
}

wireless_check_extensions() {
	# Do we have wireless extensions enabled on the interface?
	local ext=`${IWCONFIG} ${IFACE} 2>/dev/null | gawk -v IFACE=${IFACE} '{ if ($1 == IFACE) print "yes" }'`
	if [ "yes" == "${ext}" ]; then
		return 0
	fi
	return 1
}

wireless_setup_env() {
	# Get the setup for the interface
	eval essid_IFACE=\"\$\{essid_${IFACE}\}\"
	mode_IFACE=`eval echo \$\{mode_${IFACE}\} | tr '[:upper:]' '[:lower:]'`
	mode_IFACE=${mode_IFACE:-"managed"}
	eval iwconfig_IFACE=\"\$\{iwconfig_${IFACE}\}\"
	eval iwpriv_IFACE=\"\$\{iwpriv_${IFACE}\}\"
	scan_mode_IFACE=`eval echo \$\{scan_mode_${IFACE}\} | tr '[:upper:]' '[:lower:]'`
	eval sleep_scan_IFACE=\"\$\{sleep_scan_${IFACE}\}\"
	sleep_scan_IFACE=${sleep_scan_IFACE:-"0"}
	eval sleep_associate_IFACE=\"\$\{sleep_associate_${IFACE}\}\"
	sleep_associate_IFACE=${sleep_associate_IFACE:-"10"}
	associate_test_IFACE=`eval echo \$\{associate_test_${IFACE}\} | tr '[:upper:]' '[:lower:]'`
	associate_test_IFACE=${associate_test_IFACE:-"mac"}
	eval iwpriv_scan_pre_IFACE=\"\$\{iwpriv_scan_pre_${IFACE}\}\"
	eval iwpriv_scan_post_IFACE=\"\$\{iwpriv_scan_post_${IFACE}\}\"
	eval adhoc_essid_IFACE=\"\$\{adhoc_essid_${IFACE}\}\"
	eval adhoc_channel_IFACE=\"\$\{adhoc_channel_${IFACE}\}\"
}

wireless_wep_status() {
	wep_status=`${IWCONFIG} ${IFACE} 2>/dev/null | gawk '/Encryption key/ {if ($2 != "key:off") print "enabled"}'`
	wep_status=${wep_status:-"disabled"}
	wep_status="(WEP ${wep_status})"
}

wireless_report() {
	local channel=`${IWLIST} ${IFACE} channel 2>/dev/null | gawk '/Current/' | sed 's/^.*hannel[ :]\(.*\)[0-9]*.*$/\1/g'`
	channel=${channel//[![:digit:]]/}	

	local wep_status
	wireless_wep_status

	local essid=`${IWGETID} -s ${IFACE}`
	essid=${essid//\\\\/\\\\}

	local mode=`${IWGETID} -m ${IFACE} | gawk -F: '{print $2}' | tr '[:upper:]' '[:lower:]'`
	local m="to"
	if [ "ad-hoc" == "${mode}" ]; then
		m="as"
	fi
	
	einfo "    ${IFACE} connected ${m} \"${essid}\" in ${mode} mode" 
	einfo "    on channel ${channel} ${wep_status}"
}

wireless_setup_essid_env() {
	# Sets specific ESSID options
	# Replace invalid essid characters with _ so bash works
	local essid=${ESSID//[![:word:]]/_}

	eval key_ESSID=\"\$\{key_${essid}\}\"
	# iface is just here for backwards compat. When net.eth0 stops using it, so will I
	eval iface_ESSID=\"\$\{iface_${essid}\}\"
	eval ifconfig_ESSID=\"\$\{ifconfig_${essid}\}\"
	eval dhcpcd_ESSID=\"\$\{dhcpcd_${essid}\}\"
	eval routes_ESSID=\"\$\{routes_${essid}\}\"
	eval inet6_ESSID=\"\$\{inet6_${essid}\}\"
	eval gateway_ESSID=\"\$\{gateway_${essid}\}\"
	eval ifconfig_fallback_ESSID=\"\$\{ifconfig_fallback_${essid}\}\"
	eval preassociate_ESSID=\"\$\{preassociate_${essid}\}\"
	eval nameserver_ESSID=\"\$\{nameserver_${essid}\}\"
	eval domain_ESSID=\"\$\{domain_${essid}\}\"

	essid_has_config="no"
	if [ -n "${key_ESSID}" ] || [ -n "${ifconfig_IFACE}" ] || [ -n "${iface_ESSID}" ]; then
		essid_has_config="yes"
	fi

	# We set the key to default to "off" here so we can
	# tell if we have config for the ESSID or not
	key_ESSID=${key_ESSID:-"off"}

	display_ESSID=${ESSID//\\\\/\\\\}
}

wireless_setup_iface() {
	# Iface is here for backwards compat only
	if [ -n "${iface_ESSID}" ]; then
		eval "iface_${IFACE}=\"${iface_ESSID}\""
	fi
	if [ -n "${ifconfig_ESSID}" ]; then
		eval "ifconfig_${IFACE}=\"${ifconfig_ESSID}\""
	fi
	if [ -n "${dhcpcd_ESSID}" ]; then
		eval "dhcpcd_${IFACE}=\"${dhcpcd_ESSID}\""
	fi
	if [ -n "${routes_ESSID}" ]; then
		eval "routes_${IFACE}=\"${routes_ESSID}\""
	fi
	if [ -n "${inet6_ESSID}" ]; then
		eval "inet6_${IFACE}=\"${inet6_ESSID}\""
	fi
	if [ -n "${gateway_ESSID}" ]; then
		gateway="${IFACE}/${gateway_ESSID}"
	fi
	if [ -n "${ifconfig_fallback_ESSID}" ]; then
		eval "ifconfig_fallback_${IFACE}=\"${ifconfig_fallback_ESSID}\""
	fi
	if [ -n "${nameserver_ESSID}" ]; then
		echo "nameserver ${nameserver_ESSID}" > /etc/resolv.conf
		if [ -n "${domain_ESSID}" ]; then
			echo "domain ${domain_ESSID}" >> /etc/resolv.conf
			echo "search ${domain_ESSID}" >> /etc/resolv.conf
		fi
	fi
}

wireless_associate_mac() {
	local mac=`${IWCONFIG} ${IFACE} 2>/dev/null | gawk '/\<..:..:..:..:..:..\>/' | sed 's/^.*\<\(..:..:..:..:..:..\)\>.*$/\1/g'`
	if [ "FF:FF:FF:FF:FF:FF" != "${mac}" ] \
	&& [ "44:44:44:44:44:44" != "${mac}" ] \
	&& [ "00:00:00:00:00:00" != "${mac}" ]
	then
		return 0
	fi
	return 1
}

wireless_associate_quality() {
	local quality=`cat /proc/net/wireless | gawk -v IFACE=${IFACE}: '{ if ($1 == IFACE) print $3 }'`
	if [ "0" != "${quality}" ] && [ "0." != "${quality}" ]; then
		return 0
	fi
	return 1
} 

wireless_wait_for_association() {
	local counter=0
	# Some drivers don't set MAC to a bogus value when assocation is lost/fails
	# whereas they do set link quality to 0
	while ((${counter} < (${sleep_associate_IFACE}))); do
		if [ "mac" == "${associate_test_IFACE}" ]; then
			wireless_associate_mac && return 0
		elif [ "quality" == "${associate_test_IFACE}" ]; then
			wireless_associate_quality && return 0
		else
			wireless_associate_mac && wireless_associate_quality && return 0
		fi
		sleep 1
		counter=$((${counter} + 1))
	done
	return 1
}

wireless_associate() {
	local wep_required=$1
	local e
	wep_required=${wep_required:-"off"}

	if [ "${ESSID}" == "any" ]; then
		${IWCONFIG} ${IFACE} ap any 2>/dev/null
		display_ESSID="any"
	else
		wireless_setup_essid_env
		if [ "yes" == "${wep_required}" ] && [ "off" == "${key_ESSID}" ]; then
			ewarn "  WEP key is not set for \"${display_ESSID}\" - not connecting"
			return 1
		fi
		if [ "no" == "${wep_required}" ] && [ "off" != "${key_ESSID}" ]; then
			key_ESSID="off"
			ewarn "  \"${display_ESSID}\" is not WEP enabled - ignoring setting"
		fi
		e=`${IWCONFIG} ${IFACE} key ${key_ESSID} 2>&1`
		if [ -n "${e}" ] && [ "${key_ESSID}" != "off" ]; then
			local k=${ESSID//[![:word:]]/_}
			ewarn "  ${IFACE} does not support setting keys"
			ewarn "  or the parameter \"key_${k}\" is incorrect"
		fi
	fi

	e=`${IWCONFIG} ${IFACE} essid "${ESSID}" 2>&1`
	if [ -n "${e}" ] && [ "${ESSID}" != "any" ]; then
		ewarn "  ${IFACE} does not support setting ESSID to \"${display_ESSID}\""
	fi

	${IFCONFIG} ${IFACE} up 2>${devnull}
	if [ -n "${preassociate_ESSID}" ]; then
		# We have a custom script to run before associating
		ebegin "  Running an assocation setup script for \"${display_ESSID}\""
		${preassociate_ESSID} || {
			retval=$?
			eend ${retval} "Failed to run the setup script for \"${display_ESSID}\""
			return ${retval}
		}
		eend 0
	fi

	local wep_status
	wireless_wep_status
	ebegin "  Connecting to \"${display_ESSID}\" ${wep_status}"
	wireless_wait_for_association || {
		eend 1
		return 1
	}
	eend 0

	if [ "any" == "${ESSID}" ]; then
		ESSID=`${IWCONFIG} ${IFACE} | gawk -F\" '/ESSID/ {print $2}'`
		wireless_associate
		return $?
	fi
	
	wireless_setup_iface
	wireless_report
	return 0
}

wireless_setup_adhoc() {
	if [ -z "${ESSID}" ]; then
		ESSID=${adhoc_essid_IFACE}
		if [ -z "${ESSID}" ]; then
			eerror "${IFACE} requires an ESSID to be set to operate in Ad-Hoc mode"
			return 1
		fi
	fi

	local e=`${IWCONFIG} ${IFACE} mode ad-hoc 2>&1`
	if [ -n "${e}" ]; then
		ewarn "  ${IFACE} does not support setting the mode to \"Ad-Hoc\""
		return 1
	fi

	if [ -n "${adhoc_channel_IFACE}" ]; then
		e=`${IWCONFIG} ${IFACE} channel ${adhoc_channel_IFACE} 2>&1`
		if [ -n "${e}" ]; then
			ewarn "  ${IFACE} does not support setting the channel to \"${adhoc_channel_IFACE}\""
			return 1
		fi
	else
		# We default the Ad-Hoc channel to 3
		adhoc_channel_IFACE=3
		${IWCONFIG} ${IFACE} channel ${adhoc_channel_IFACE} 2>/dev/null
	fi

	wireless_setup_essid_env
	wireless_setup_iface
                                                                                
	e=`${IWCONFIG} ${IFACE} key ${key_ESSID} 2>&1`
	if [ -n "${e}" ] && [ "${key_ESSID}" != "off" ]; then
		ewarn "  ${IFACE} does not support setting the key ${key_IFACE}"
	fi

	e=`${IWCONFIG} ${IFACE} essid "${ESSID}" 2>&1`
	if [ -n "${e}" ]; then
		error "  ${IFACE} does not support setting ESSID to \"${display_ESSID}\""
		return 1
	fi

	wireless_report
	return 0
}

wireless_apply_defaults() {
	# Set some defaults
	${IWCONFIG} ${IFACE} rts off &>/dev/null
	${IWCONFIG} ${IFACE} frag off &>/dev/null
	${IWCONFIG} ${IFACE} power off &>/dev/null
	${IWCONFIG} ${IFACE} txpower auto &>/dev/null
	${IWCONFIG} ${IFACE} key off &>/dev/null
	${IWCONFIG} ${IFACE} essid any &>/dev/null
}

wireless_scan_aps() {
	local e
	# First, we may need to change mode to scan in
	if [ -n "${scan_mode_IFACE}" ] && [ "${scan_mode_IFACE}" != "${mode_IFACE}" ]; then
		e=`${IWCONFIG} ${IFACE} mode ${scan_mode_IFACE} 2>&1`
		if [ -n "${e}" ]; then
			ewarn "  ${IFACE} does not support setting the mode to \"${scan_mode_IFACE}\""
		fi
	fi

	# Next we set any private driver ioctls needed
	if [ -n "${iwpriv_scan_pre_IFACE}" ]; then
		e=`${IWPRIV} ${IFACE} ${iwpriv_scan_pre_IFACE} 2>&1`
		if [ -n "${e}" ]; then
			ewarn "  ${IFACE} does not support the following private ioctls" \
			ewarn "    \"${iwpriv_scan_pre_IFACE}\""
		fi
	fi

	einfo "  Scanning for access points"
	${IFCONFIG} ${IFACE} up >${devnull}
	sleep ${sleep_scan_IFACE}
	local scan_IFACE=`${IWLIST} ${IFACE} scan 2>/dev/null`

	# We may need to unset the previous private driver ioctls
	if [ -n "${iwpriv_scan_post_IFACE}" ]; then
		e=`${IWPRIV} ${IFACE} ${iwpriv_scan_post_IFACE} 2>&1`
		if [ -n "${e}" ]; then
			ewarn "  ${IFACE} does not support the following private ioctls" \
			ewarn "    \"${iwpriv_scan_post_IFACE}\""
		fi
	fi

	# Change back to the desired mode
	if [ -n "${scan_mode_IFACE}" ] && [ "${scan_mode_IFACE}" != "${mode_IFACE}" ]; then
		if [ -n "${mode_IFACE}" ]; then
			${IWCONFIG} ${IFACE} mode ${mode_IFACE} 2>${devnull}
		else
			${IWCONFIG} ${IFACE} mode Managed 2>${devnull}
		fi
	fi

	if [ -z "${scan_IFACE}" ]; then
		if [ -z "${preferred_aps}" ] && [ -z "${adhoc_essid_IFACE}" ]; then
			eerror "${IFACE} does not support scanning!"
			eerror "You need to configure an ESSID in /etc/conf.d/wireless"
			eerror "    essid_${IFACE}=\"ESSID\""
			eerror "    (set to \"any\" to pick the strongest available access point)"
			eerror "or configure preferred_aps to force a list of access points"
			eerror "or configure defaulting to Ad-Hoc when Managed fails"
			return 1
		else
			ewarn "  ${IFACE} does not support scanning"
			return 0
		fi
	fi

	local val
	local i=-1
	local IFS=$'\n'
#	scan_IFACE=`echo ${scan_IFACE} | grep -E 'Address:|ESSID:|Encryption key:'`
	for val in ${scan_IFACE}; do
		local t=`echo ${val} | gawk -F: '/ESSID/ {print $2}'`
		if [ ${#t} -gt 2 ]; then
			t=${t#*\"}
			essid_APs[i]=${t%*\"}
		else
			t=`echo ${val} | gawk '/Address/ {print}' | sed 's/^.*\<\(..:..:..:..:..:..\)\>.*$/\1/g'`
			if [ ${#t} -gt 0 ]; then
				i=$((${i} + 1))
				mac_APs[i]=${t}
			else
				t=`echo ${val} | gawk -F: '/Encryption key/ {print $2}'`
				if [ ${#t} -gt 0 ]; then
					if [ "off" == "${t}" ]; then
						enc_APs[i]="no";
					else
						enc_APs[i]="yes";
					fi
				fi
			fi
		fi
	done
}

wireless_start() {
	${IFCONFIG} ${IFACE} down >/dev/null
	wireless_apply_defaults

	# Apply the user configuration
	local e
        if [ -n "${iwconfig_IFACE}" ]; then
                local e=`${IWCONFIG} ${IFACE} ${iwconfig_IFACE} 2>&1`
		if [ -n "${e}" ]; then
			ewarn "  ${IFACE} does not support the following configuration commands"
			ewarn "    \"${iwconfig_IFACE}\""
		fi
        fi
        if [ -n "${iwpriv_IFACE}" ]; then
		local e=`${IWPRIV} ${IFACE} ${iwpriv_IFACE} 2>&1`
		if [ -n "${e}" ]; then
			ewarn "  ${IFACE} does not support the following private ioctls"
			ewarn "    \"${iwpriv_IFACE}\""
		fi
        fi

	# Set the mode. We do this after the config as the mode could have been changed there
	if [ -n "${mode_IFACE}" ]; then
		e=`${IWCONFIG} ${IFACE} mode ${mode_IFACE} 2>&1`
		if [ -n "${e}" ]; then
                	ewarn "  ${IFACE} does not support setting the mode to \"${mode_IFACE}\""
		fi
	else
		${IWCONFIG} ${IFACE} mode Managed 2>/dev/null
	fi

	# Setup ad-hoc mode?
	if [ "ad-hoc" == "${mode_IFACE}" ]; then
		ESSID=${essid_IFACE}
		wireless_setup_adhoc
		return $?
	fi

	if [ "managed" != "${mode_IFACE}" ]; then
		eerror "Only managed and ad-hoc modes are supported"
		return 1
	fi

	# Ensure that we have a working association test
	if [ "mac" != "${associate_test_IFACE}" ] \
	&& [ "quality" != "${associate_test_IFACE}" ] \
	&& [ "all" != "${associate_test_IFACE}" ]
	then
		eerror "associate_test_${IFACE} can only be \"mac\", \"quality\" or \"all\""
		return 1
	fi

	# Has an ESSID been forced?
	if [ -n "${essid_IFACE}" ]; then
		ESSID=${essid_IFACE};
		wireless_associate && return 0
		if [ -n "${adhoc_essid_IFACE}" ]; then
			ESSID=${adhoc_essid_IFACE}
			wireless_setup_adhoc
			return $?
		fi
		if [ "${essid_IFACE}" != "any" ]; then
			return 1;
		fi
	fi

	local i j first_AP first_AP_enc
	local -a mac_APs essid_APs enc_APs
	if [ "${essid_IFACE}" != "any" ]; then
		wireless_scan_aps || return 1

		for ((i = 0; i < ${#mac_APs[@]}; i = i + 1)); do
			local keyreq=""
			if [ "yes" == "${enc_APs[i]}" ]; then
				keyreq="(WEP required)"
			fi
			if [ -z "${essid_APs[i]}" ]; then
				einfo "    Found ${mac_APs[i]} ${keyreq}"
			else
				einfo "    Found \"${essid_APs[i]//\\\\/\\\\}\" at ${mac_APs[i]} ${keyreq}"
			fi
			local mac_essid
			eval mac_essid=\"\$\{essid_${mac_APs[i]//:/}\}\"
			if [ -n "${mac_essid}" ]; then
				essid_APs[i]=${mac_essid}
				einfo "      mapping to \"${essid_APs[i]//\\\\/\\\\}\""
			fi
		done

		# Check for preferred APs
		for ((i = 0; i < ${#preferred_aps[@]}; i = i + 1)); do
			ESSID=${preferred_aps[i]}
			for ((j = 0; j < ${#mac_APs[@]}; j = j + 1)); do
				if [ "${ESSID}" == "${essid_APs[j]}" ]; then
					wireless_associate ${enc_APs[j]} && return 0
					break
				fi
			done
		done

		# Check for configured APs that are NOT in our preferred list and try to connect
		for ((i = 0; i < ${#mac_APs[@]}; i = i + 1)); do
			ESSID=${essid_APs[i]}
			local has_preferred="no"
			for ((j = 0; j < ${#preferred_aps[@]}; j = j + 1)); do
				if [ "${ESSID}" == "${preferred_aps[j]}" ]; then
					has_preferred="yes"
					break
				fi
			done
			if [ "no" == "${has_preferred}" ]; then
				wireless_setup_essid_env
				if [ "yes" == "${essid_has_config}" ]; then
					wireless_associate ${enc_APs[i]} && return 0
				elif [ -z "${first_AP}" ] && [ -n "${ESSID}" ]; then
					first_AP=${ESSID}
					first_AP_enc=${enc_APs[i]}
				fi
			fi
		done
	fi

	if [ -n "${preferred_aps}" ]; then
		ewarn "  No preferred access points found"
		ewarn "  Trying to force preferred incase they are hidden"
		for ((i = 0; i < ${#preferred_aps[@]}; i = i + 1)); do
			ESSID=${preferred_aps[i]}
			local found_AP="no"
			for ((j = 0; j < ${#mac_APs[@]}; j = j + 1)); do
				if [ "${ESSID}" == "${essid_APs[j]}" ]; then
					found_AP="yes"
					break
				fi
			done
			if [ "no" == "${found_AP}" ]; then
				wireless_associate && return 0
			fi
		done
		ewarn "  Failed to associate with any preferred access points"
	fi

	e="Couldn't associate with any access points"
	if [ -n "${first_AP}" ]; then
		ESSID=${first_AP}
		wireless_associate ${first_AP_enc} && return 0
	elif [ 0 -eq ${#mac_APs[@]} ]; then
		e="No access points found for ${IFACE}"
	fi
	
	if [ -n "${adhoc_essid_IFACE}" ]; then
		ewarn "  ${e}"
		ewarn "  Defaulting to Ad-Hoc"
		ESSID=${adhoc_essid_IFACE}
		wireless_setup_adhoc
		return $?
	fi
	
	eerror ${e}
	return 1
}

wireless_up() {
	local IFACE="${1}"
	
	local IFCONFIG IWCONFIG IWLIST IWPRIV IWGETID
	local essid_IFACE mode_IFACE config_IFACE priv_IFACE
	local scan_mode_IFACE sleep_scan_IFACE sleep_associate_IFACE
	local associate_test_IFACE
	local priv_scan_pre_IFACE priv_scan_post_IFACE
	local adhoc_essid_IFACE adhoc_channel_IFACE

	local essid_has_config key_ESSID iface_ESSID ifconfig_ESSID dhcpcd_ESSID
	local routes_ESSID inet6_ESSID ifconfig_fallback_ESSID
	local gateway_ESSID nameserver_ESSID domain_ESSID preassociate_ESSID
	local display_ESSID

	#Setup IFS incase parent script has modified it
	local IFS=$' '$'\n'$'\t'

	wireless_setup_programs
	wireless_checkconfig || return 1

	wireless_check_extensions || {
		einfo "  Wireless extensions not found for ${IFACE}"
		return 0
	}

	wireless_setup_env

	local retval=0
	einfo "  Configuring wireless network for ${IFACE}"
	wireless_start || {
		retval=$?
		eerror "Failed to configure wireless for ${IFACE}"
	}

	# We take the interface back down even if we connect OK otherwise
	# DHCP won't work
	${IFCONFIG} ${IFACE} down

	return ${retval}
}

wireless_down() {
	local IFACE="${1}"

	local IFCONFIG IWCONFIG IWLIST IWPRIV IWGETID

	#Setup IFS incase parent script os modified it
	local IFS=$' '$'\n'$'\t'

	wireless_setup_programs
	wireless_checkconfig || return 1

        wireless_check_extensions || {
                einfo "  Wireless extensions not found for ${IFACE}"
                return 0
        }
	
	ebegin "  Removing wireless configuration from ${IFACE}"
	wireless_apply_defaults
	eend 0
}
